@@ -1,5 +1,4 @@
-"""
-This module holds the code of the Markdown extension responsible for matching "autodoc" instructions.
+"""This module holds the code of the Markdown extension responsible for matching "autodoc" instructions.
 
 The extension is composed of a Markdown [block processor](https://python-markdown.github.io/extensions/api/#blockparser)
 that matches indented blocks starting with a line like '::: identifier'.
@@ -24,8 +23,8 @@ instruction:
 """
 import re
 from collections import ChainMap
-from typing import Any, Mapping, MutableSequence, Tuple
-from xml.etree.ElementTree import XML, Element, ParseError  # noqa: S405 (we choose to trust the XML input)
+from typing import Mapping, MutableSequence, Sequence, Tuple
+from xml.etree.ElementTree import Element
 
 import yaml
 from jinja2.exceptions import TemplateNotFound
@@ -33,58 +32,23 @@ from markdown import Markdown
 from markdown.blockparser import BlockParser
 from markdown.blockprocessors import BlockProcessor
 from markdown.extensions import Extension
-from markdown.util import AtomicString
+from markdown.treeprocessors import Treeprocessor
+from mkdocs_autorefs.plugin import AutorefsPlugin
 
-from mkdocstrings.handlers.base import CollectionError, Handlers
+from mkdocstrings.handlers.base import CollectionError, CollectorItem, Handlers
 from mkdocstrings.loggers import get_logger
-from mkdocstrings.references import AutoRefInlineProcessor
+
+try:
+    from mkdocs.exceptions import PluginError  # New in MkDocs 1.2
+except ImportError:
+    PluginError = SystemExit
+
 
 log = get_logger(__name__)
 
-ENTITIES = """
-    <!DOCTYPE html [
-        <!ENTITY hellip '&amp;hellip;'>
-        <!ENTITY laquo '&amp;laquo;'>
-        <!ENTITY ldquo '&amp;ldquo;'>
-        <!ENTITY lsquo '&amp;lsquo;'>
-        <!ENTITY mdash '&amp;mdash;'>
-        <!ENTITY nbsp '&amp;nbsp;'>
-        <!ENTITY ndash '&amp;ndash;'>
-        <!ENTITY para '&amp;para;'>
-        <!ENTITY raquo '&amp;raquo;'>
-        <!ENTITY rdquo '&amp;rdquo;'>
-        <!ENTITY rsquo '&amp;rsquo;'>
-    ]>
-"""
-
-
-def atomic_brute_cast(tree: Element) -> Element:
-    """
-    Cast every node's text into an atomic string to prevent further processing on it.
-
-    Since we generate the final HTML with Jinja templates, we do not want other inline or tree processors
-    to keep modifying the data, so this function is used to mark the complete tree as "do not touch".
-
-    Reference: issue [Python-Markdown/markdown#920](https://github.com/Python-Markdown/markdown/issues/920).
-
-    On a side note: isn't `atomic_brute_cast` such a beautiful function name?
-
-    Arguments:
-        tree: An XML node, used like the root of an XML tree.
-
-    Returns:
-        The same node, recursively modified by side-effect. You can skip re-assigning the return value.
-    """
-    if tree.text:
-        tree.text = AtomicString(tree.text)
-    for child in tree:
-        atomic_brute_cast(child)
-    return tree
-
 
 class AutoDocProcessor(BlockProcessor):
-    """
-    Our "autodoc" Markdown block processor.
+    """Our "autodoc" Markdown block processor.
 
     It has a [`test` method][mkdocstrings.extension.AutoDocProcessor.test] that tells if a block matches a criterion,
     and a [`run` method][mkdocstrings.extension.AutoDocProcessor.run] that processes it.
@@ -95,9 +59,10 @@ class AutoDocProcessor(BlockProcessor):
 
     regex = re.compile(r"^(?P<heading>#{1,6} *|)::: ?(?P<name>.+?) *$", flags=re.MULTILINE)
 
-    def __init__(self, parser: BlockParser, md: Markdown, config: dict, handlers: Handlers) -> None:
-        """
-        Initialize the object.
+    def __init__(
+        self, parser: BlockParser, md: Markdown, config: dict, handlers: Handlers, autorefs: AutorefsPlugin
+    ) -> None:
+        """Initialize the object.
 
         Arguments:
             parser: A `markdown.blockparser.BlockParser` instance.
@@ -105,16 +70,17 @@ class AutoDocProcessor(BlockProcessor):
             config: The [configuration][mkdocstrings.plugin.MkdocstringsPlugin.config_scheme]
                 of the `mkdocstrings` plugin.
             handlers: A [mkdocstrings.handlers.base.Handlers][] instance.
+            autorefs: A [mkdocs_autorefs.plugin.AutorefsPlugin][] instance.
         """
         super().__init__(parser=parser)
         self.md = md
         self._config = config
         self._handlers = handlers
+        self._autorefs = autorefs
         self._updated_env = False
 
     def test(self, parent: Element, block: str) -> bool:
-        """
-        Match our autodoc instructions.
+        """Match our autodoc instructions.
 
         Arguments:
             parent: The parent element in the XML tree.
@@ -126,8 +92,7 @@ class AutoDocProcessor(BlockProcessor):
         return bool(self.regex.search(block))
 
     def run(self, parent: Element, blocks: MutableSequence[str]) -> None:
-        """
-        Run code on the matched blocks.
+        """Run code on the matched blocks.
 
         The identifier and configuration lines are retrieved from a matched block
         and used to collect and render an object.
@@ -151,8 +116,18 @@ class AutoDocProcessor(BlockProcessor):
             identifier = match["name"]
             heading_level = match["heading"].count("#")
             log.debug(f"Matched '::: {identifier}'")
-            xml_element = self.process_block(identifier, block, heading_level)
-            parent.append(xml_element)
+
+            html, headings = self._process_block(identifier, block, heading_level)
+            el = Element("div", {"class": "mkdocstrings"})
+            # The final HTML is inserted as opaque to subsequent processing, and only revealed at the end.
+            el.text = self.md.htmlStash.store(html)
+            # So we need to duplicate the headings directly (and delete later), just so 'toc' can pick them up.
+            el.extend(headings)
+
+            for heading in headings:
+                self._autorefs.register_anchor(self._autorefs.current_page, heading.attrib["id"])
+
+            parent.append(el)
 
         if the_rest:
             # This block contained unindented line(s) after the first indented
@@ -160,9 +135,8 @@ class AutoDocProcessor(BlockProcessor):
             # list for future processing.
             blocks.insert(0, the_rest)
 
-    def process_block(self, identifier: str, yaml_block: str, heading_level: int = 0) -> Element:
-        """
-        Process an autodoc block.
+    def _process_block(self, identifier: str, yaml_block: str, heading_level: int = 0) -> Tuple[str, Sequence[Element]]:
+        """Process an autodoc block.
 
         Arguments:
             identifier: The identifier of the object to collect and render.
@@ -170,12 +144,11 @@ class AutoDocProcessor(BlockProcessor):
             heading_level: Suggested level of the the heading to insert (0 to ignore).
 
         Raises:
-            CollectionError: When something wrong happened during collection.
-            ParseError: When the generated HTML could not be parsed as XML.
+            PluginError: When something wrong happened during collection.
             TemplateNotFound: When a template used for rendering could not be found.
 
         Returns:
-            A new XML element.
+            Rendered HTML and the list of heading elements encoutered.
         """
         config = yaml.safe_load(yaml_block) or {}
         handler_name = self._handlers.get_handler_name(config)
@@ -190,14 +163,16 @@ class AutoDocProcessor(BlockProcessor):
 
         log.debug("Collecting data")
         try:
-            data: Any = handler.collector.collect(identifier, selection)
-        except CollectionError:
-            log.error(f"Could not collect '{identifier}'")
-            raise
+            data: CollectorItem = handler.collector.collect(identifier, selection)
+        except CollectionError as exception:
+            log.error(str(exception))
+            if PluginError is SystemExit:  # When MkDocs 1.2 is sufficiently common, this can be dropped.
+                log.error(f"Error reading page '{self._autorefs.current_page}':")
+            raise PluginError(f"Could not collect '{identifier}'") from exception
 
         if not self._updated_env:
             log.debug("Updating renderer's env")
-            handler.renderer.update_env(self.md, self._config)
+            handler.renderer._update_env(self.md, self._config)  # noqa: W0212 (protected member OK)
             self._updated_env = True
 
         log.debug("Rendering templates")
@@ -210,20 +185,11 @@ class AutoDocProcessor(BlockProcessor):
             )
             raise
 
-        log.debug("Loading HTML back into XML tree")
-        rendered = ENTITIES + rendered
-        try:
-            xml_contents = XML(rendered)
-        except ParseError as error:
-            log_xml_parse_error(str(error), rendered)
-            raise
-
-        return atomic_brute_cast(xml_contents)  # type: ignore
+        return (rendered, handler.renderer.get_headings())
 
 
 def get_item_configs(handler_config: dict, config: dict) -> Tuple[Mapping, Mapping]:
-    """
-    Get the selection and rendering configuration merged into the global configuration of the given handler.
+    """Get the selection and rendering configuration merged into the global configuration of the given handler.
 
     Arguments:
         handler_config: The global configuration of a handler. It can be an empty dictionary.
@@ -237,74 +203,57 @@ def get_item_configs(handler_config: dict, config: dict) -> Tuple[Mapping, Mappi
     return item_selection_config, item_rendering_config
 
 
-def log_xml_parse_error(error: str, xml_text: str) -> None:
-    """
-    Log an XML parsing error.
-
-    If the error is a tag mismatch, augment the log message.
-
-    Arguments:
-        error: The error message (no traceback).
-        xml_text: The XML text that generated the parsing error.
-    """
-    message = error
-    mismatched_tag = "mismatched tag" in error
-    undefined_entity = "undefined entity" in error
-
-    if mismatched_tag or undefined_entity:
-        line_column = error[error.rfind(":") + 1 :]
-        line, column = line_column.split(", ")
-        lineno = int(line[line.rfind(" ") + 1 :])
-        columnno = int(column[column.rfind(" ") + 1 :])
-
-        line = xml_text.split("\n")[lineno - 1]
-        if mismatched_tag:
-            character = line[columnno]
-            message += (
-                f" (character {character}):\n{line}\n"
-                "If your Markdown contains angle brackets < >, try to wrap them between backticks `< >`, "
-                "or replace them with &lt; and &gt;"
-            )
-        elif undefined_entity:
-            message += f":\n{line}\n"
-    log.error(message)
+class _PostProcessor(Treeprocessor):
+    def run(self, root: Element):
+        carry_text = ""
+        for el in reversed(root):  # Reversed mainly for the ability to mutate during iteration.
+            if el.tag == "div" and el.get("class") == "mkdocstrings":
+                # Delete the duplicated headings along with their container, but keep the text (i.e. the actual HTML).
+                carry_text = (el.text or "") + carry_text
+                root.remove(el)
+            elif carry_text:
+                el.tail = (el.tail or "") + carry_text
+                carry_text = ""
+        if carry_text:
+            root.text = (root.text or "") + carry_text
 
 
 class MkdocstringsExtension(Extension):
-    """
-    Our Markdown extension.
+    """Our Markdown extension.
 
     It cannot work outside of `mkdocstrings`.
     """
 
-    blockprocessor_priority = 75  # Right before markdown.blockprocessors.HashHeaderProcessor
-    inlineprocessor_priority = 168  # Right after markdown.inlinepatterns.ReferenceInlineProcessor
-
-    def __init__(self, config: dict, handlers: Handlers, **kwargs) -> None:
-        """
-        Initialize the object.
+    def __init__(self, config: dict, handlers: Handlers, autorefs: AutorefsPlugin, **kwargs) -> None:
+        """Initialize the object.
 
         Arguments:
             config: The configuration items from `mkdocs` and `mkdocstrings` that must be passed to the block processor
                 when instantiated in [`extendMarkdown`][mkdocstrings.extension.MkdocstringsExtension.extendMarkdown].
             handlers: A [mkdocstrings.handlers.base.Handlers][] instance.
+            autorefs: A [mkdocs_autorefs.plugin.AutorefsPlugin][] instance.
             kwargs: Keyword arguments used by `markdown.extensions.Extension`.
         """
         super().__init__(**kwargs)
         self._config = config
         self._handlers = handlers
+        self._autorefs = autorefs
 
     def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802 (casing: parent method's name)
-        """
-        Register the extension.
+        """Register the extension.
 
         Add an instance of our [`AutoDocProcessor`][mkdocstrings.extension.AutoDocProcessor] to the Markdown parser.
 
         Arguments:
             md: A `markdown.Markdown` instance.
         """
-        md.registerExtension(self)
-        processor = AutoDocProcessor(md.parser, md, self._config, self._handlers)
-        md.parser.blockprocessors.register(processor, "mkdocstrings", self.blockprocessor_priority)
-        ref_processor = AutoRefInlineProcessor(md)
-        md.inlinePatterns.register(ref_processor, "mkdocstrings", self.inlineprocessor_priority)
+        md.parser.blockprocessors.register(
+            AutoDocProcessor(md.parser, md, self._config, self._handlers, self._autorefs),
+            "mkdocstrings",
+            priority=75,  # Right before markdown.blockprocessors.HashHeaderProcessor
+        )
+        md.treeprocessors.register(
+            _PostProcessor(md.parser),
+            "mkdocstrings_post",
+            priority=4,  # Right after 'toc'.
+        )
