@@ -1,5 +1,4 @@
-"""
-Base module for handlers.
+"""Base module for handlers.
 
 This module contains the base classes for implementing collectors, renderers, and the combination of the two: handlers.
 
@@ -9,25 +8,26 @@ It also provides two methods:
 - `teardown`, that will teardown all the cached handlers, and then clear the cache.
 """
 
-import functools
 import importlib
-import re
-import textwrap
 from abc import ABC, abstractmethod
 from pathlib import Path
-from typing import Any, Dict, Optional, Sequence
-from xml.etree.ElementTree import Element  # noqa: S405 (we choose to trust the XML input)
+from typing import Any, Dict, Iterable, Optional, Sequence
+from xml.etree.ElementTree import Element, tostring
 
 from jinja2 import Environment, FileSystemLoader
 from markdown import Markdown
-from markdown.extensions import Extension
-from markdown.treeprocessors import Treeprocessor
 from markupsafe import Markup
-from pymdownx.highlight import Highlight
 
+from mkdocstrings.handlers.rendering import (
+    HeadingShiftingTreeprocessor,
+    Highlighter,
+    IdPrependingTreeprocessor,
+    MkdocstringsInnerExtension,
+)
 from mkdocstrings.loggers import get_template_logger
 
-handlers_cache: Dict[str, Any] = {}
+CollectorItem = Any
+
 TEMPLATES_DIR = Path(__file__).parent.parent / "templates"
 
 
@@ -39,80 +39,8 @@ class ThemeNotSupported(Exception):
     """An exception raised to tell a theme is not supported."""
 
 
-def do_highlight(
-    src: str,
-    guess_lang: bool = False,
-    language: str = None,
-    inline: bool = False,
-    dedent: bool = True,
-    line_nums: bool = False,
-    line_start: int = 1,
-) -> str:
-    """
-    Highlight a code-snippet.
-
-    This function is used as a filter in Jinja templates.
-
-    Arguments:
-        src: The code to highlight.
-        guess_lang: Whether to guess the language or not.
-        language: Explicitly tell what language to use for highlighting.
-        inline: Whether to do inline highlighting.
-        dedent: Whether to dedent the code before highlighting it or not.
-        line_nums: Whether to add line numbers in the result.
-        line_start: The line number to start with.
-
-    Returns:
-        The highlighted code as HTML text, marked safe (not escaped for HTML).
-    """
-    if dedent:
-        src = textwrap.dedent(src)
-
-    highlighter = Highlight(use_pygments=True, guess_lang=guess_lang, linenums=line_nums)
-    result = highlighter.highlight(src=src, language=language, linestart=line_start, inline=inline)
-
-    if inline:
-        return Markup(f'<code class="highlight language-{language}">{result.text}</code>')
-    return Markup(result)
-
-
-def do_js_highlight(
-    src: str,
-    guess_lang: bool = False,  # noqa: W0613 (we must accept the same parameters as do_highlight)
-    language: str = None,
-    inline: bool = False,
-    dedent: bool = True,
-    line_nums: bool = False,  # noqa: W0613
-    line_start: int = 1,  # noqa: W0613
-) -> str:
-    """
-    Prepare a code-snippet for JS highlighting.
-
-    This function is used as a filter in Jinja templates.
-
-    Arguments:
-        src: The code to highlight.
-        guess_lang: Whether to guess the language or not.
-        language: Explicitly tell what language to use for highlighting.
-        inline: Whether to do inline highlighting.
-        dedent: Whether to dedent the code before highlighting it or not.
-        line_nums: Whether to add line numbers in the result.
-        line_start: The line number to start with.
-
-    Returns:
-        The code properly wrapped for later highlighting by JavaScript.
-    """
-    if dedent:
-        src = textwrap.dedent(src)
-    if inline:
-        src = re.sub(r"\n\s*", "", src)
-        return Markup(f'<code class="highlight">{src}</code>')
-    return Markup(f'<div class="highlight {language or ""}"><pre><code>\n{src}\n</code></pre></div>')
-
-
 def do_any(seq: Sequence, attribute: str = None) -> bool:
-    """
-    Check if at least one of the item in the sequence evaluates to true.
+    """Check if at least one of the item in the sequence evaluates to true.
 
     The `any` builtin as a filter for Jinja templates.
 
@@ -128,32 +56,8 @@ def do_any(seq: Sequence, attribute: str = None) -> bool:
     return any(_[attribute] for _ in seq)
 
 
-def do_convert_markdown(md: Markdown, text: str, heading_level: int, html_id: str = "") -> Markup:
-    """
-    Render Markdown text; for use inside templates.
-
-    Arguments:
-        md: A `markdown.Markdown` instance.
-        text: The text to convert.
-        heading_level: The base heading level to start all Markdown headings from.
-        html_id: The HTML id of the element that's considered the parent of this element.
-
-    Returns:
-        An HTML string.
-    """
-    md.treeprocessors["mkdocstrings_headings"].shift_by = heading_level
-    md.treeprocessors["mkdocstrings_ids"].id_prefix = html_id and html_id + "--"
-    try:
-        return Markup(md.convert(text))
-    finally:
-        md.treeprocessors["mkdocstrings_headings"].shift_by = 0
-        md.treeprocessors["mkdocstrings_ids"].id_prefix = ""
-        md.reset()
-
-
 class BaseRenderer(ABC):
-    """
-    The base renderer class.
+    """The base renderer class.
 
     Inherit from this class to implement a renderer.
 
@@ -161,14 +65,15 @@ class BaseRenderer(ABC):
     You can also override the `update_env` method, to add more filters to the Jinja environment,
     making them available in your Jinja templates.
 
-    To define a fallback theme, add a `FALLBACK_THEME` class-variable.
+    To define a fallback theme, add a `fallback_theme` class-variable.
+    To add custom CSS, add an `extra_css` variable or create an 'style.css' file beside the templates.
     """
 
     fallback_theme: str = ""
+    extra_css = ""
 
     def __init__(self, directory: str, theme: str, custom_templates: Optional[str] = None) -> None:
-        """
-        Initialize the object.
+        """Initialize the object.
 
         If the given theme is not supported (it does not exist), it will look for a `fallback_theme` attribute
         in `self` to use as a fallback theme.
@@ -180,16 +85,22 @@ class BaseRenderer(ABC):
         """
         paths = []
 
-        if custom_templates is not None:
-            paths.append(Path(custom_templates) / directory / theme)
-
         themes_dir = TEMPLATES_DIR / directory
 
         paths.append(themes_dir / theme)
 
-        if self.fallback_theme != "":
+        if self.fallback_theme:
             paths.append(themes_dir / self.fallback_theme)
 
+        for path in paths:
+            css_path = path / "style.css"
+            if css_path.is_file():
+                self.extra_css += "\n" + css_path.read_text(encoding="utf-8")
+                break
+
+        if custom_templates is not None:
+            paths.insert(0, Path(custom_templates) / directory / theme)
+
         self.env = Environment(
             autoescape=True,
             loader=FileSystemLoader(paths),
@@ -198,17 +109,12 @@ class BaseRenderer(ABC):
         self.env.filters["any"] = do_any
         self.env.globals["log"] = get_template_logger()
 
-        if theme == "readthedocs":
-            highlight_function = do_js_highlight
-        else:
-            highlight_function = do_highlight
-
-        self.env.filters["highlight"] = highlight_function
+        self._headings = []
+        self._md = None  # To be populated in `update_env`.
 
     @abstractmethod
-    def render(self, data: Any, config: dict) -> str:
-        """
-        Render a template using provided data and configuration options.
+    def render(self, data: CollectorItem, config: dict) -> str:
+        """Render a template using provided data and configuration options.
 
         Arguments:
             data: The collected data to render.
@@ -218,33 +124,128 @@ class BaseRenderer(ABC):
             The rendered template as HTML.
         """  # noqa: DAR202 (excess return section)
 
-    def update_env(self, md: Markdown, config: dict) -> None:
+    def get_anchor(self, data: CollectorItem) -> Optional[str]:
+        """Return the canonical identifier (HTML anchor) for a collected item.
+
+        This must match what the renderer would've actually rendered,
+        e.g. if rendering the item contains `<h2 id="foo">...` then the return value should be "foo".
+
+        Arguments:
+            data: The collected data.
+
+        Returns:
+            The HTML anchor (without '#') as a string, or None if this item doesn't have an anchor.
+        """  # noqa: DAR202 (excess return section)
+
+    def do_convert_markdown(self, text: str, heading_level: int, html_id: str = "") -> Markup:
+        """Render Markdown text; for use inside templates.
+
+        Arguments:
+            text: The text to convert.
+            heading_level: The base heading level to start all Markdown headings from.
+            html_id: The HTML id of the element that's considered the parent of this element.
+
+        Returns:
+            An HTML string.
         """
-        Update the Jinja environment.
+        treeprocessors = self._md.treeprocessors
+        treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = heading_level
+        treeprocessors[IdPrependingTreeprocessor.name].id_prefix = html_id and html_id + "--"
+        try:
+            return Markup(self._md.convert(text))
+        finally:
+            treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = 0
+            treeprocessors[IdPrependingTreeprocessor.name].id_prefix = ""
+            self._md.reset()
+
+    def do_heading(
+        self,
+        content: str,
+        heading_level: int,
+        *,
+        hidden: bool = False,
+        toc_label: Optional[str] = None,
+        **attributes: str,
+    ) -> Markup:
+        """Render an HTML heading and register it for the table of contents. For use inside templates.
+
+        Arguments:
+            content: The HTML within the heading.
+            heading_level: The level of heading (e.g. 3 -> `h3`).
+            hidden: If True, only register it for the table of contents, don't render anything.
+            toc_label: The title to use in the table of contents ('data-toc-label' attribute).
+            attributes: Any extra HTML attributes of the heading.
+
+        Returns:
+            An HTML string.
+        """
+        # First, produce the "fake" heading, for ToC only.
+        el = Element(f"h{heading_level}", attributes)
+        if toc_label is None:
+            toc_label = content.unescape() if isinstance(el, Markup) else content
+        el.set("data-toc-label", toc_label)
+        self._headings.append(el)
+
+        if hidden:
+            return Markup('<a id="{0}"></a>').format(attributes["id"])
+
+        # Now produce the actual HTML to be rendered. The goal is to wrap the HTML content into a heading.
+        # Start with a heading that has just attributes (no text), and add a placeholder into it.
+        el = Element(f"h{heading_level}", attributes)
+        el.append(Element("mkdocstrings-placeholder"))
+        # Tell the 'toc' extension to make its additions if configured so.
+        toc = self._md.treeprocessors["toc"]
+        if toc.use_anchors:
+            toc.add_anchor(el, attributes["id"])
+        if toc.use_permalinks:
+            toc.add_permalink(el, attributes["id"])
+
+        # The content we received is HTML, so it can't just be inserted into the tree. We had marked the middle
+        # of the heading with a placeholder that can never occur (text can't directly contain angle brackets).
+        # Now this HTML wrapper can be "filled" by replacing the placeholder.
+        html_with_placeholder = tostring(el, encoding="unicode")
+        assert (
+            html_with_placeholder.count("<mkdocstrings-placeholder />") == 1
+        ), f"Bug in mkdocstrings: failed to replace in {html_with_placeholder!r}"
+        html = html_with_placeholder.replace("<mkdocstrings-placeholder />", content)
+        return Markup(html)
+
+    def get_headings(self) -> Sequence[Element]:
+        """Return and clear the headings gathered so far.
+
+        Returns:
+            A list of HTML elements.
+        """
+        result = list(self._headings)
+        self._headings.clear()
+        return result
+
+    def update_env(self, md: Markdown, config: dict) -> None:  # noqa: W0613 (unused argument 'config')
+        """Update the Jinja environment.
 
         Arguments:
             md: The Markdown instance. Useful to add functions able to convert Markdown into the environment filters.
             config: Configuration options for `mkdocs` and `mkdocstrings`, read from `mkdocs.yml`. See the source code
                 of [mkdocstrings.plugin.MkdocstringsPlugin.on_config][] to see what's in this dictionary.
         """
-        extensions = config["mdx"] + [ShiftHeadingsExtension(), PrefixIdsExtension()]
-        configs = dict(config["mdx_configs"])
-        # Prevent a bug that happens due to treeprocessors running on the same fragment both as the inner doc and as
-        # part of the re-integrated doc. Namely, the permalink '¶' would be appended twice. This is the only known
-        # non-idempotent effect of an extension, so specifically prevent it on the inner doc.
-        try:
-            configs["toc"] = dict(configs["toc"], permalink=False)
-        except KeyError:
-            pass
+        self._md = md
+        self.env.filters["highlight"] = Highlighter(md).highlight
+        self.env.filters["convert_markdown"] = self.do_convert_markdown
+        self.env.filters["heading"] = self.do_heading
 
-        md = Markdown(extensions=extensions, extension_configs=configs)
+    def _update_env(self, md: Markdown, config: dict):
+        extensions = config["mdx"] + [MkdocstringsInnerExtension(self._headings)]
 
-        self.env.filters["convert_markdown"] = functools.partial(do_convert_markdown, md)
+        new_md = Markdown(extensions=extensions, extension_configs=config["mdx_configs"])
+        # MkDocs adds its own (required) extension that's not part of the config. Propagate it.
+        if "relpath" in md.treeprocessors:
+            new_md.treeprocessors.register(md.treeprocessors["relpath"], "relpath", priority=0)
+
+        self.update_env(new_md, config)
 
 
 class BaseCollector(ABC):
-    """
-    The base collector class.
+    """The base collector class.
 
     Inherit from this class to implement a collector.
 
@@ -253,9 +254,8 @@ class BaseCollector(ABC):
     """
 
     @abstractmethod
-    def collect(self, identifier: str, config: dict) -> Any:
-        """
-        Collect data given an identifier and selection configuration.
+    def collect(self, identifier: str, config: dict) -> CollectorItem:
+        """Collect data given an identifier and selection configuration.
 
         In the implementation, you typically call a subprocess that returns JSON, and load that JSON again into
         a Python dictionary for example, though the implementation is completely free.
@@ -272,8 +272,7 @@ class BaseCollector(ABC):
         """  # noqa: DAR202 (excess return section)
 
     def teardown(self) -> None:
-        """
-        Teardown the collector.
+        """Teardown the collector.
 
         This method should be implemented to, for example, terminate a subprocess
         that was started when creating the collector instance.
@@ -281,8 +280,7 @@ class BaseCollector(ABC):
 
 
 class BaseHandler:
-    """
-    The base handler class.
+    """The base handler class.
 
     Inherit from this class to implement a handler.
 
@@ -290,8 +288,7 @@ class BaseHandler:
     """
 
     def __init__(self, collector: BaseCollector, renderer: BaseRenderer) -> None:
-        """
-        Initialize the object.
+        """Initialize the object.
 
         Arguments:
             collector: A collector instance.
@@ -302,16 +299,14 @@ class BaseHandler:
 
 
 class Handlers:
-    """
-    A collection of handlers.
+    """A collection of handlers.
 
     Do not instantiate this directly. [The plugin][mkdocstrings.plugin.MkdocstringsPlugin] will keep one instance of
     this for the purpose of caching. Use [mkdocstrings.plugin.MkdocstringsPlugin.get_handler][] for convenient access.
     """
 
     def __init__(self, config: dict) -> None:
-        """
-        Initialize the object.
+        """Initialize the object.
 
         Arguments:
             config: Configuration options for `mkdocs` and `mkdocstrings`, read from `mkdocs.yml`. See the source code
@@ -320,9 +315,27 @@ class Handlers:
         self._config = config
         self._handlers: Dict[str, BaseHandler] = {}
 
-    def get_handler_name(self, config: dict) -> str:
+    def get_anchor(self, identifier: str) -> Optional[str]:
+        """Return the canonical HTML anchor for the identifier, if any of the seen handlers can collect it.
+
+        Arguments:
+            identifier: The identifier (one that [collect][mkdocstrings.handlers.base.BaseCollector.collect] can accept).
+
+        Returns:
+            A string - anchor without '#', or None if there isn't any identifier familiar with it.
         """
-        Return the handler name defined in an "autodoc" instruction YAML configuration, or the global default handler.
+        for handler in self._handlers.values():
+            try:
+                anchor = handler.renderer.get_anchor(handler.collector.collect(identifier, {}))
+            except CollectionError:
+                continue
+            else:
+                if anchor is not None:
+                    return anchor
+        return None
+
+    def get_handler_name(self, config: dict) -> str:
+        """Return the handler name defined in an "autodoc" instruction YAML configuration, or the global default handler.
 
         Arguments:
             config: A configuration dictionary, obtained from YAML below the "autodoc" instruction.
@@ -336,8 +349,7 @@ class Handlers:
         return config["default_handler"]
 
     def get_handler_config(self, name: str) -> dict:
-        """
-        Return the global configuration of the given handler.
+        """Return the global configuration of the given handler.
 
         Arguments:
             name: The name of the handler to get the global configuration of.
@@ -351,8 +363,7 @@ class Handlers:
         return {}
 
     def get_handler(self, name: str, handler_config: Optional[dict] = None) -> BaseHandler:
-        """
-        Get a handler thanks to its name.
+        """Get a handler thanks to its name.
 
         This function dynamically imports a module named "mkdocstrings.handlers.NAME", calls its
         `get_handler` method to get an instance of a handler, and caches it in dictionary.
@@ -378,91 +389,18 @@ class Handlers:
             )  # type: ignore
         return self._handlers[name]
 
-    def teardown(self):
+    @property
+    def seen_handlers(self) -> Iterable[BaseHandler]:
+        """Get the handlers that were encountered so far throughout the build.
+
+        Returns:
+            An iterable of instances of [`BaseHandler`][mkdocstrings.handlers.base.BaseHandler]
+            (usable only to loop through it).
+        """
+        return self._handlers.values()
+
+    def teardown(self) -> None:
         """Teardown all cached handlers and clear the cache."""
-        for handler in self._handlers.values():
+        for handler in self.seen_handlers:
             handler.collector.teardown()
         self._handlers.clear()
-
-
-class _IdPrependingTreeprocessor(Treeprocessor):
-    def __init__(self, md, id_prefix: str):
-        super().__init__(md)
-        self.id_prefix = id_prefix
-
-    def run(self, root: Element):
-        if not self.id_prefix:
-            return
-        for el in root.iter():
-            id_attr = el.get("id")
-            if id_attr:
-                el.set("id", self.id_prefix + id_attr)
-
-            href_attr = el.get("href")
-            if href_attr and href_attr.startswith("#"):
-                el.set("href", "#" + self.id_prefix + href_attr[1:])
-
-            name_attr = el.get("name")
-            if name_attr:
-                el.set("name", self.id_prefix + name_attr)
-
-            if el.tag == "label":
-                for_attr = el.get("for")
-                if for_attr:
-                    el.set("for", self.id_prefix + for_attr)
-
-
-class PrefixIdsExtension(Extension):
-    """Prepend the configured prefix to IDs of all HTML elements."""
-
-    treeprocessor_priority = 4  # Right after 'toc' (needed because that extension adds ids to headers).
-
-    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802 (casing: parent method's name)
-        """
-        Register the extension, with a treeprocessor under the name 'mkdocstrings_ids'.
-
-        Arguments:
-            md: A `markdown.Markdown` instance.
-        """
-        md.registerExtension(self)
-        md.treeprocessors.register(
-            _IdPrependingTreeprocessor(md, ""),
-            "mkdocstrings_ids",
-            self.treeprocessor_priority,
-        )
-
-
-class _HeadingShiftingTreeprocessor(Treeprocessor):
-    def __init__(self, md, shift_by: int):
-        super().__init__(md)
-        self.shift_by = shift_by
-
-    def run(self, root: Element):
-        if not self.shift_by:
-            return
-        for el in root.iter():
-            match = re.fullmatch(r"([Hh])([1-6])", el.tag)
-            if match:
-                level = int(match[2]) + self.shift_by
-                level = max(1, min(level, 6))
-                el.tag = f"{match[1]}{level}"
-
-
-class ShiftHeadingsExtension(Extension):
-    """Shift levels of all Markdown headings according to the configured base level."""
-
-    treeprocessor_priority = 12
-
-    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802 (casing: parent method's name)
-        """
-        Register the extension, with a treeprocessor under the name 'mkdocstrings_headings'.
-
-        Arguments:
-            md: A `markdown.Markdown` instance.
-        """
-        md.registerExtension(self)
-        md.treeprocessors.register(
-            _HeadingShiftingTreeprocessor(md, 0),
-            "mkdocstrings_headings",
-            self.treeprocessor_priority,
-        )
