@@ -1,5 +1,4 @@
-"""
-This module implements a handler for the Python language.
+"""This module implements a handler for the Python language.
 
 The handler collects data with [`pytkdocs`](https://github.com/pawamoy/pytkdocs).
 """
@@ -7,21 +6,21 @@ The handler collects data with [`pytkdocs`](https://github.com/pawamoy/pytkdocs)
 import json
 import os
 import sys
+import traceback
 from collections import ChainMap
 from subprocess import PIPE, Popen  # noqa: S404 (what other option, more secure that PIPE do we have? sockets?)
 from typing import Any, List, Optional
 
 from markdown import Markdown
 
-from mkdocstrings.handlers.base import BaseCollector, BaseHandler, BaseRenderer, CollectionError
+from mkdocstrings.handlers.base import BaseCollector, BaseHandler, BaseRenderer, CollectionError, CollectorItem
 from mkdocstrings.loggers import get_logger
 
 log = get_logger(__name__)
 
 
 class PythonRenderer(BaseRenderer):
-    """
-    The class responsible for loading Jinja templates and rendering them.
+    """The class responsible for loading Jinja templates and rendering them.
 
     It defines some configuration options, implements the `render` method,
     and overrides the `update_env` method of the [`BaseRenderer` class][mkdocstrings.handlers.base.BaseRenderer].
@@ -47,8 +46,7 @@ class PythonRenderer(BaseRenderer):
         "group_by_category": True,
         "heading_level": 2,
     }
-    """
-    The default rendering options.
+    """The default rendering options.
 
     Option | Type | Description | Default
     ------ | ---- | ----------- | -------
@@ -65,7 +63,7 @@ class PythonRenderer(BaseRenderer):
     **`heading_level`** | `int` | The initial heading level to use. | `2`
     """  # noqa: E501
 
-    def render(self, data: Any, config: dict) -> str:  # noqa: D102 (ignore missing docstring)
+    def render(self, data: CollectorItem, config: dict) -> str:  # noqa: D102 (ignore missing docstring)
         final_config = ChainMap(config, self.default_config)
 
         template = self.env.get_template(f"{data['category']}.html")
@@ -79,6 +77,9 @@ class PythonRenderer(BaseRenderer):
             **{"config": final_config, data["category"]: data, "heading_level": heading_level, "root": True},
         )
 
+    def get_anchor(self, data: CollectorItem) -> str:  # noqa: D102 (ignore missing docstring)
+        return data.get("path")
+
     def update_env(self, md: Markdown, config: dict) -> None:  # noqa: D102 (ignore missing docstring)
         super().update_env(md, config)
         self.env.trim_blocks = True
@@ -87,16 +88,14 @@ class PythonRenderer(BaseRenderer):
 
 
 class PythonCollector(BaseCollector):
-    """
-    The class responsible for loading Jinja templates and rendering them.
+    """The class responsible for loading Jinja templates and rendering them.
 
     It defines some configuration options, implements the `render` method,
     and overrides the `update_env` method of the [`BaseRenderer` class][mkdocstrings.handlers.base.BaseRenderer].
     """
 
     default_config: dict = {"filters": ["!^_[^_]"]}
-    """
-    The default selection options.
+    """The default selection options.
 
     Option | Type | Description | Default
     ------ | ---- | ----------- | -------
@@ -119,8 +118,7 @@ class PythonCollector(BaseCollector):
     """
 
     def __init__(self, setup_commands: Optional[List[str]] = None) -> None:
-        """
-        Initialize the object.
+        """Initialize the object.
 
         When instantiating a Python collector, we open a subprocess in the background with `subprocess.Popen`.
         It will allow us to feed input to and read output from this subprocess, keeping it alive during
@@ -161,9 +159,8 @@ class PythonCollector(BaseCollector):
             env=env,
         )
 
-    def collect(self, identifier: str, config: dict) -> Any:
-        """
-        Collect the documentation tree given an identifier and selection options.
+    def collect(self, identifier: str, config: dict) -> CollectorItem:
+        """Collect the documentation tree given an identifier and selection options.
 
         In this method, we feed one line of JSON to the standard input of the subprocess that was opened
         during instantiation of the collector. Then we read one line of JSON on its standard output.
@@ -208,15 +205,13 @@ class PythonCollector(BaseCollector):
         try:
             result = json.loads(stdout)
         except json.decoder.JSONDecodeError as exception:
-            log.error(f"Error while loading JSON: {stdout}")
-            raise CollectionError(str(exception)) from exception
+            error = "\n".join(("Error while loading JSON:", stdout, traceback.format_exc()))
+            raise CollectionError(error) from exception
 
         error = result.get("error")
         if error:
-            message = f"Collection failed: {error}"
             if "traceback" in result:
-                message += f"\n{result['traceback']}"
-            log.error(message)
+                error += f"\n{result['traceback']}"
             raise CollectionError(error)
 
         for loading_error in result["loading_errors"]:
@@ -250,8 +245,7 @@ def get_handler(
     setup_commands: Optional[List[str]] = None,
     **config: Any,
 ) -> PythonHandler:
-    """
-    Simply return an instance of `PythonHandler`.
+    """Simply return an instance of `PythonHandler`.
 
     Arguments:
         theme: The theme to use when rendering contents.
@@ -269,8 +263,7 @@ def get_handler(
 
 
 def rebuild_category_lists(obj: dict) -> None:
-    """
-    Recursively rebuild the category lists of a collected object.
+    """Recursively rebuild the category lists of a collected object.
 
     Since `pytkdocs` dumps JSON on standard output, it must serialize the object-tree and flatten it to reduce data
     duplication and avoid cycle-references. Indeed, each node of the object-tree has a `children` list, containing
